@@ -28,6 +28,8 @@ import processing.app.Preferences;
 import processing.app.Sketch;
 import processing.app.SketchCode;
 import processing.core.*;
+import processing.app.I18n;
+import static processing.app.I18n._;
 
 import java.io.*;
 import java.util.*;
@@ -36,14 +38,15 @@ import java.util.zip.*;
 
 public class Compiler implements MessageConsumer {
   static final String BUGS_URL =
-    "http://code.google.com/p/arduino/issues/list";
+    _("http://github.com/arduino/Arduino/issues");
   static final String SUPER_BADNESS =
-    "Compiler error, please submit this code to " + BUGS_URL;
+    I18n.format(_("Compiler error, please submit this code to {0}"), BUGS_URL);
 
   Sketch sketch;
   String buildPath;
   String primaryClassName;
   boolean verbose;
+  boolean sketchIsCompiled;
 
   RunnerException exception;
 
@@ -66,6 +69,7 @@ public class Compiler implements MessageConsumer {
     this.buildPath = buildPath;
     this.primaryClassName = primaryClassName;
     this.verbose = verbose;
+    this.sketchIsCompiled = false;
 
     // the pms object isn't used for anything but storage
     MessageStream pms = new MessageStream(this);
@@ -74,7 +78,7 @@ public class Compiler implements MessageConsumer {
     Map<String, String> boardPreferences = Base.getBoardPreferences();
     String core = boardPreferences.get("build.core");
     if (core == null) {
-    	RunnerException re = new RunnerException("No board selected; please choose a board from the Tools > Board menu.");
+    	RunnerException re = new RunnerException(_("No board selected; please choose a board from the Tools > Board menu."));
       re.hideStackTrace();
       throw re;
     }
@@ -128,6 +132,7 @@ public class Compiler implements MessageConsumer {
                findFilesInPath(buildPath, "c", false),
                findFilesInPath(buildPath, "cpp", false),
                boardPreferences));
+   sketchIsCompiled = true;
 
    // 2. compile the libraries, outputting .o files to: <buildPath>/<library>/
 
@@ -183,12 +188,18 @@ public class Compiler implements MessageConsumer {
    }
 
     // 4. link it all together into the .elf file
-
+    // For atmega2560, need --relax linker option to link larger
+    // programs correctly.
+    String optRelax = "";
+    String atmega2560 = new String ("atmega2560");
+    if ( atmega2560.equals(boardPreferences.get("build.mcu")) ) {
+        optRelax = new String(",--relax");
+    }
    sketch.setCompilingProgress(60);
     List baseCommandLinker = new ArrayList(Arrays.asList(new String[] {
       avrBasePath + "avr-gcc",
       "-Os",
-      "-Wl,--gc-sections",
+      "-Wl,--gc-sections"+optRelax,
       "-mmcu=" + boardPreferences.get("build.mcu"),
       "-o",
       buildPath + File.separator + primaryClassName + ".elf"
@@ -261,7 +272,11 @@ public class Compiler implements MessageConsumer {
  		
     for (File file : cSources) {
         String objectPath = buildPath + File.separator + file.getName() + ".o";
-        objectPaths.add(new File(objectPath));
+        String dependPath = buildPath + File.separator + file.getName() + ".d";
+        File objectFile = new File(objectPath);
+        File dependFile = new File(dependPath);
+        objectPaths.add(objectFile);
+        if (is_already_compiled(file, objectFile, dependFile, boardPreferences)) continue;
         execAsynchronously(getCommandCompilerC(avrBasePath, includePaths,
                                                file.getAbsolutePath(),
                                                objectPath,
@@ -270,7 +285,11 @@ public class Compiler implements MessageConsumer {
 
     for (File file : cppSources) {
         String objectPath = buildPath + File.separator + file.getName() + ".o";
-        objectPaths.add(new File(objectPath));
+        String dependPath = buildPath + File.separator + file.getName() + ".d";
+        File objectFile = new File(objectPath);
+        File dependFile = new File(dependPath);
+        objectPaths.add(objectFile);
+        if (is_already_compiled(file, objectFile, dependFile, boardPreferences)) continue;
         execAsynchronously(getCommandCompilerCPP(avrBasePath, includePaths,
                                                  file.getAbsolutePath(),
                                                  objectPath,
@@ -280,6 +299,68 @@ public class Compiler implements MessageConsumer {
     return objectPaths;
   }
 
+  private boolean is_already_compiled(File src, File obj, File dep, Map<String, String> prefs) {
+    boolean ret=true;
+    try {
+      //System.out.println("\n  is_already_compiled: begin checks: " + obj.getPath());
+      if (!obj.exists()) return false;  // object file (.o) does not exist
+      if (!dep.exists()) return false;  // dep file (.d) does not exist
+      long src_modified = src.lastModified();
+      long obj_modified = obj.lastModified();
+      if (src_modified >= obj_modified) return false;  // source modified since object compiled
+      if (src_modified >= dep.lastModified()) return false;  // src modified since dep compiled
+      BufferedReader reader = new BufferedReader(new FileReader(dep.getPath()));
+      String line;
+      boolean need_obj_parse = true;
+      while ((line = reader.readLine()) != null) {
+        if (line.endsWith("\\")) {
+          line = line.substring(0, line.length() - 1);
+        }
+        line = line.trim();
+        if (line.length() == 0) continue; // ignore blank lines
+        if (need_obj_parse) {
+          // line is supposed to be the object file - make sure it really is!
+          if (line.endsWith(":")) {
+            line = line.substring(0, line.length() - 1);
+            String objpath = obj.getCanonicalPath();
+            File linefile = new File(line);
+            String linepath = linefile.getCanonicalPath();
+            //System.out.println("  is_already_compiled: obj =  " + objpath);
+            //System.out.println("  is_already_compiled: line = " + linepath);
+            if (objpath.compareTo(linepath) == 0) {
+              need_obj_parse = false;
+              continue;
+            } else {
+              ret = false;  // object named inside .d file is not the correct file!
+              break;
+            }
+          } else {
+            ret = false;  // object file supposed to end with ':', but didn't
+            break;
+          }
+        } else {
+          // line is a prerequisite file
+          File prereq = new File(line);
+          if (!prereq.exists()) {
+            ret = false;  // prerequisite file did not exist
+            break;
+          }
+          if (prereq.lastModified() >= obj_modified) {
+            ret = false;  // prerequisite modified since object was compiled
+            break;
+          }
+          //System.out.println("  is_already_compiled:  prerequisite ok");
+        }
+      }
+      reader.close();
+    } catch (Exception e) {
+      return false;  // any error reading dep file = recompile it
+    }
+    if (ret && (verbose || Preferences.getBoolean("build.verbose"))) {
+      System.out.println("  Using previously compiled: " + obj.getPath());
+    }
+    return ret;
+  }
 
   boolean firstErrorFound;
   boolean secondErrorFound;
@@ -320,10 +401,8 @@ public class Compiler implements MessageConsumer {
     boolean compiling = true;
     while (compiling) {
       try {
-        if (in.thread != null)
-          in.thread.join();
-        if (err.thread != null)
-          err.thread.join();
+        in.join();
+        err.join();
         result = process.waitFor();
         //System.out.println("result is " + result);
         compiling = false;
@@ -340,11 +419,12 @@ public class Compiler implements MessageConsumer {
 
     if (result > 1) {
       // a failure in the tool (e.g. unable to locate a sub-executable)
-      System.err.println(command[0] + " returned " + result);
+      System.err.println(
+	  I18n.format(_("{0} returned {1}"), command[0], result));
     }
 
     if (result != 0) {
-      RunnerException re = new RunnerException("Error compiling.");
+      RunnerException re = new RunnerException(_("Error compiling."));
       re.hideStackTrace();
       throw re;
     }
@@ -383,55 +463,71 @@ public class Compiler implements MessageConsumer {
       String error = pieces[3], msg = "";
       
       if (pieces[3].trim().equals("SPI.h: No such file or directory")) {
-        error = "Please import the SPI library from the Sketch > Import Library menu.";
-        msg = "\nAs of Arduino 0019, the Ethernet library depends on the SPI library." +
-              "\nYou appear to be using it or another library that depends on the SPI library.\n\n";
+        error = _("Please import the SPI library from the Sketch > Import Library menu.");
+        msg = _("\nAs of Arduino 0019, the Ethernet library depends on the SPI library." +
+              "\nYou appear to be using it or another library that depends on the SPI library.\n\n");
       }
       
       if (pieces[3].trim().equals("'BYTE' was not declared in this scope")) {
-        error = "The 'BYTE' keyword is no longer supported.";
-        msg = "\nAs of Arduino 1.0, the 'BYTE' keyword is no longer supported." +
-              "\nPlease use Serial.write() instead.\n\n";
+        error = _("The 'BYTE' keyword is no longer supported.");
+        msg = _("\nAs of Arduino 1.0, the 'BYTE' keyword is no longer supported." +
+              "\nPlease use Serial.write() instead.\n\n");
       }
       
       if (pieces[3].trim().equals("no matching function for call to 'Server::Server(int)'")) {
-        error = "The Server class has been renamed EthernetServer.";
-        msg = "\nAs of Arduino 1.0, the Server class in the Ethernet library " +
-              "has been renamed to EthernetServer.\n\n";
+        error = _("The Server class has been renamed EthernetServer.");
+        msg = _("\nAs of Arduino 1.0, the Server class in the Ethernet library " +
+              "has been renamed to EthernetServer.\n\n");
       }
       
       if (pieces[3].trim().equals("no matching function for call to 'Client::Client(byte [4], int)'")) {
-        error = "The Client class has been renamed EthernetClient.";
-        msg = "\nAs of Arduino 1.0, the Client class in the Ethernet library " +
-              "has been renamed to EthernetClient.\n\n";
+        error = _("The Client class has been renamed EthernetClient.");
+        msg = _("\nAs of Arduino 1.0, the Client class in the Ethernet library " +
+              "has been renamed to EthernetClient.\n\n");
       }
       
       if (pieces[3].trim().equals("'Udp' was not declared in this scope")) {
-        error = "The Udp class has been renamed EthernetUdp.";
-        msg = "\nAs of Arduino 1.0, the Udp class in the Ethernet library " +
-              "has been renamed to EthernetClient.\n\n";
+        error = _("The Udp class has been renamed EthernetUdp.");
+        msg = _("\nAs of Arduino 1.0, the Udp class in the Ethernet library " +
+              "has been renamed to EthernetUdp.\n\n");
       }
       
       if (pieces[3].trim().equals("'class TwoWire' has no member named 'send'")) {
-        error = "Wire.send() has been renamed Wire.write().";
-        msg = "\nAs of Arduino 1.0, the Wire.send() function was renamed " +
-              "to Wire.write() for consistency with other libraries.\n\n";
+        error = _("Wire.send() has been renamed Wire.write().");
+        msg = _("\nAs of Arduino 1.0, the Wire.send() function was renamed " +
+              "to Wire.write() for consistency with other libraries.\n\n");
       }
       
       if (pieces[3].trim().equals("'class TwoWire' has no member named 'receive'")) {
-        error = "Wire.receive() has been renamed Wire.read().";
-        msg = "\nAs of Arduino 1.0, the Wire.receive() function was renamed " +
-              "to Wire.read() for consistency with other libraries.\n\n";
+        error = _("Wire.receive() has been renamed Wire.read().");
+        msg = _("\nAs of Arduino 1.0, the Wire.receive() function was renamed " +
+              "to Wire.read() for consistency with other libraries.\n\n");
       }
 
-      RunnerException e = sketch.placeException(error, pieces[1], PApplet.parseInt(pieces[2]) - 1);
+      if (pieces[3].trim().equals("'Mouse' was not declared in this scope")) {
+        error = _("'Mouse' only supported on the Arduino Leonardo");
+        //msg = _("\nThe 'Mouse' class is only supported on the Arduino Leonardo.\n\n");
+      }
+      
+      if (pieces[3].trim().equals("'Keyboard' was not declared in this scope")) {
+        error = _("'Keyboard' only supported on the Arduino Leonardo");
+        //msg = _("\nThe 'Keyboard' class is only supported on the Arduino Leonardo.\n\n");
+      }
+      
+      RunnerException e = null;
+      if (!sketchIsCompiled) {
+        // Place errors when compiling the sketch, but never while compiling libraries
+        // or the core.  The user's sketch might contain the same filename!
+        e = sketch.placeException(error, pieces[1], PApplet.parseInt(pieces[2]) - 1);
+      }
 
       // replace full file path with the name of the sketch tab (unless we're
       // in verbose mode, in which case don't modify the compiler output)
       if (e != null && !verbose) {
         SketchCode code = sketch.getCode(e.getCodeIndex());
-        String fileName = code.isExtension(sketch.getDefaultExtension()) ? code.getPrettyName() : code.getFileName();
-        s = fileName + ":" + e.getCodeLine() + ": error: " + pieces[3] + msg;        
+        String fileName = (code.isExtension("ino") || code.isExtension("pde")) ? code.getPrettyName() : code.getFileName();
+        int lineNum = e.getCodeLine() + 1;
+        s = fileName + ":" + lineNum + ": error: " + pieces[3] + msg;        
       }
             
       if (exception == null && e != null) {
@@ -453,8 +549,10 @@ public class Compiler implements MessageConsumer {
       "-g", // include debugging info (so errors include line numbers)
       "-assembler-with-cpp",
       "-mmcu=" + boardPreferences.get("build.mcu"),
-      "-DF_CPU=" + boardPreferences.get("build.f_cpu"),
+      "-DF_CPU=" + boardPreferences.get("build.f_cpu"),      
       "-DARDUINO=" + Base.REVISION,
+      "-DUSB_VID=" + boardPreferences.get("build.vid"),
+      "-DUSB_PID=" + boardPreferences.get("build.pid"),
     }));
 
     for (int i = 0; i < includePaths.size(); i++) {
@@ -481,7 +579,10 @@ public class Compiler implements MessageConsumer {
       "-fdata-sections",
       "-mmcu=" + boardPreferences.get("build.mcu"),
       "-DF_CPU=" + boardPreferences.get("build.f_cpu"),
-      "-DARDUINO=" + Base.REVISION,
+      "-MMD", // output dependancy info
+      "-DUSB_VID=" + boardPreferences.get("build.vid"),
+      "-DUSB_PID=" + boardPreferences.get("build.pid"),
+      "-DARDUINO=" + Base.REVISION, 
     }));
 		
     for (int i = 0; i < includePaths.size(); i++) {
@@ -489,7 +590,8 @@ public class Compiler implements MessageConsumer {
     }
 
     baseCommandCompiler.add(sourceName);
-    baseCommandCompiler.add("-o"+ objectName);
+    baseCommandCompiler.add("-o");
+    baseCommandCompiler.add(objectName);
 
     return baseCommandCompiler;
   }
@@ -510,6 +612,9 @@ public class Compiler implements MessageConsumer {
       "-fdata-sections",
       "-mmcu=" + boardPreferences.get("build.mcu"),
       "-DF_CPU=" + boardPreferences.get("build.f_cpu"),
+      "-MMD", // output dependancy info
+      "-DUSB_VID=" + boardPreferences.get("build.vid"),
+      "-DUSB_PID=" + boardPreferences.get("build.pid"),      
       "-DARDUINO=" + Base.REVISION,
     }));
 
@@ -518,7 +623,8 @@ public class Compiler implements MessageConsumer {
     }
 
     baseCommandCompilerCPP.add(sourceName);
-    baseCommandCompilerCPP.add("-o"+ objectName);
+    baseCommandCompilerCPP.add("-o");
+    baseCommandCompilerCPP.add(objectName);
 
     return baseCommandCompilerCPP;
   }
